Defined.enable!

class ThingTank

  class Character < CouchRest::Model::Base

    include ThingTank::SharedMethods

    property :update_me

    before_validation do |character|
      couchrest_attribute_will_change!('update_me') # or update_me_will_change! will also do
      character.changed_to_character_hash. each do |k,v|
        if v.is_a?(ThingTank)
          if v.id
            self[k] = v.to_sym
          else
            self[k] = v.to_character_hash(true)
          end
        # only do this for [ThingTank.new, ThingTank.new]
        # all of them either having an id or not (just the first entry is checked)
        elsif v && v.is_a?(Array) && !v.empty? && v.first.is_a?(ThingTank)
          if v.first.id
            self[k] = v.collect do |tt|
              tt.to_sym
            end  
          else
            self[k] = v.collect do |tt|
              tt.to_character_hash(true)
            end
          end
        end
      end
    end

    class << self
      # we need this hack with after_inherited and defined or active record is going wild, see: http://stackoverflow.com/questions/790626/ruby-can-i-have-something-like-classinherited-thats-triggered-only-after-the
      # 'I'm trying to add behavior to activerecord models, but I need all the model customizations to go through before I mess with it. I'm trying to add behavior to activerecord models, but I need all the model customizations to go through before I mess with it. '
      # exactly my case
      # simple inherited leads to chaos here
      def after_inherited(child)
        @__sub_classes ||= {}
        @__sub_classes[child.to_s.to_sym] = child
      end

      def defined(*args)
        superclass.after_inherited(self) if superclass.respond_to?(:after_inherited)
      end

      def __subclasses
        @__sub_classes ||= {}
      end

      def property(name, *args)
        @character_properties ||= []
        @character_properties << name.to_s
        #p [:prop, self.name, @character_properties]
        super
      end

      def character_properties
        if superclass.respond_to? :character_properties
          (superclass.character_properties || []).concat(@character_properties || []).uniq
        else
          @character_properties
        end
      end

      def -(key)
        [self,key]
      end

      def get(id, db = database)
        doc = ThingTank.get(id)
        return nil if doc.nil?
        return doc.to_character(self)
      end

      def get!(id, db = database)
        doc = ThingTank.get!(id)
        doc.to_character(self)
      end

      def wants(*modules)
        @wishes ||= []
        @wishes = @wishes.concat modules
      end

      def wishes
        @wishes
      end

      def design
        raise "design is not supported in ThingTank::Character, please use the 'character_view' method in a ThingTank subclass design definition"
      end

      def view_by(*args)
        raise "view_by is not supported in ThingTank::Character, please use the 'character_view_by' method in a ThingTank subclass"
      end

    end

    def -(key)
      [self,key]
    end

    def _doc
      if database.is_a?(FakeBase) 
        database._doc
      else
        nil
      end
    end

    # the virtual _doc that contains me, you should not need it normally
    def _character_doc
      if database.is_a?(FakeBase) 
        database._character_doc
      else
        nil
      end
    end

    # use to_doc to put characters changes back to the doc without saving the doc
    def to_doc
      if changed?
        valid = valid?
        changed_to_character_hash().each do |k,v|
          _character_doc[k] = v
        end
        if valid
          @changed_attributes.clear if @changed_attributes
        else
          _doc.errors.add self.class.to_s, self.errors.messages
        end
      end
    end

    def flush_to_doc
      if changed?
        to_doc
        _character_doc.save
      end
    end

    # use reload to get the latest properties from the doc without reloading all from the database
    def reload
      attrs = _character_doc.as(self.class).to_character_hash
      prepare_all_attributes(attrs, :directly_set_attributes => true)
      @changed_attributes.clear if @changed_attributes
      self
    end

    # use reload! to get the latest properties from the database, the doc will be reloaded and the character to
    def reload!
      _character_doc.reload
      reload
    end

    def to_character(klass, key, &code)
      _character_doc.to_character(klass, key, &code)
    end

    def first(key)
      _character_doc.first(key)
    end

    def last(key)
      _character_doc.last(key)
    end

    def add_character(klass, key=nil, &code)
      _character_doc.add_character(klass, key, &code)
    end
  end
end
