class ThingTank 

  # tracks the dependencies of a ThingTank, such as saving of characters, children ThingTanks, registration of characters
  class Dependencies
    def initialize(doc)
      @doc = doc
      @registration = []
      @save = []
      @children = {}
      #@character_objects = []
      #@just_reloaded = []
    end

    attr_reader :parent

    def inspect
      { :registration => @registration, :save => @save, :children => @children }.inspect
    end

    def register_characters
      @registration.each { |character| @doc.register_character(character) }
    end

    # we don't do this. there is no proper way to update all given characters
    # one must not rely on a character properties after the doc has been manipulated
    # use Character#reload to get the latest from the doc object to the character and Character#reload! to reload the doc from the database and reload the character then
    #def add_character_object(obj)
    #  @character_objects << obj
    #end

    # we don't do this. there is no proper way to update all given characters
    # one must not rely on a character properties after the doc has been manipulated
    # use Character#reload to get the latest from the doc object to the character and Character#reload! to reload the doc from the database and reload the character then
    #def reload_character_objects()
    #  @just_reloaded = []
    #  @character_objects.each do |character|
    #    unless @just_reloaded.include?(character)
    #      @just_reloaded << character
    #      character.reload
    #    end
    #  end
    #end

    def save_all()
      refresh true
      @save.each { |character_instance| character_instance.save }
    end

    def add_character(character)
      @registration << character
    end

    def remove_character(character)
      @registration.delete(character) if has_character?(character)
    end

    def has_character?(klass)
      @registration.include? klass
    end

    def save_character(character_instance)
      @save << character_instance unless @save.include? character_instance
    end

    def already_saved(character_instance)
      @save.delete character_instance
    end

    def refresh(save=false, with_parent=true)
      register_characters
      @children.each do |key,child|
        @doc[key] = case child
        when Array
          child.collect do |d|
            d.dependencies.refresh(save, false)
            d.save if save
            d.to_character_hash
          end
        else
          child.dependencies.refresh(save, false)
          child.save if save
          child.to_character_hash
        end
      end
      refresh_parent() if with_parent
    end

    def refresh_parent() 
      @parent.dependencies.refresh if @parent
    end
   
    def add_child(key,child)
      raise "something wrong here #{key} #{child.inspect}" unless child.is_a?(ThingTank) || child.is_a?(Array)
      if child.is_a? Array
        child.each do |c|
          c.dependencies.set_parent(@doc)
        end
      else
        child.dependencies.set_parent(@doc)
      end      
      @children[key] = child
    end

    def find_root_doc()
      return @doc unless @parent
      _d = @doc
      doc = _d
      while doc.dependencies.parent do
        doc = doc.dependencies.parent
      end
      return doc
    end

    protected

    def set_parent(parent)
      @parent = parent
    end
  end
end
