# a replacement database for characters
class ThingTank::FakeBase

  def initialize(doc, character, add_dep = true)
    @db = doc.database
    @doc = doc
    @character = character
    @doc.dependencies.add_character(character) if add_dep
  end

  def method_missing(meth, *args)
    raise "don't call #{self.class}: #{meth.inspect}(#{args.inspect})"
  end

  def is_a?(klass)
    return true if klass == CouchRest::Database
    super
  end

  def to_character(klass)
    self.class.new(@doc, klass).get()
  end

  def result(ok=true)
    {"ok" => ok}
  end

  def save_doc(character_doc)
    if character_doc.is_a? ThingTank # I dunno why this happens
      @doc.dependencies.refresh_parent()
    else
      @doc.save_character_attributes(character_doc) if character_doc.changed?
    end
    result
  end

  def delete_doc(*character_docs)    
    character_docs.each do |character_doc|
      # delete only the attributes from me that are no part of another character, since we will have different doc ids
      # for each doc we will need to identify the remaining attributes
      doc, save = (character_doc["_id"].nil? || character_doc["_id"] == @doc.id) ?
        [@doc, false] : 
        [@db.get(character_doc["_id"]), true]

      doc.delete_character(character_doc)
      doc.save if save # if all characters are from our doc, only save once: at the end
    end
    @doc.save if @doc.changed?
    result()
  end

  def get(id=:doc_id)
    id = @doc.id if id == :doc_id
    doc = id == @doc.id ? @doc : @db.get(id)
    doc.get_character(@character, self)
  end

  def _character_doc
    @doc
  end

  def _doc
    @doc._root
  end
end

