class ClassValidator < ActiveModel::EachValidator
  def validate_single_class(record, attribute, values, options)
    if options[:in] # validates :hosting_account, :class => [Hash, String]
      valid = false
      options[:in].each do |klass|
        if values.is_a?(klass)
          valid = true
        end
      end
      unless valid
        record.errors.add attribute, "invalid class is #{values.class.to_s}, should be one of #{options[:in].join(', ')}"
      end
    else # validates :hosting_account, :class => Hash
      unless values.is_a?(options[:with])
        record.errors.add attribute, "invalid class is #{values.class.to_s}, should be #{options[:with]}"
      end
    end
  end


  def validate_each(record, attribute, values)
    case values
    when NilClass
      return nil
    when Array
      if options[:with] && options[:with] == Array
        validate_single_class(record, attribute, values, options)
      else
        values.each do |value|
          validate_single_class(record, attribute, value, options)
        end
      end
    else
      validate_single_class(record, attribute, values, options)
    end
  end
end


class CharacterValidator < ActiveModel::EachValidator
  def validate_single_character(record, attribute, values, options)
    if options[:in] # validates :hosting_account, :character => [HostingAccount, ...]
      valid = false
      options[:in].each do |character|
        character = character.new(values)
        if character.valid?
          valid = true
        end
      end
      unless valid
        record.errors.add attribute, "invalid character is #{values.inspect}, should be one of #{options[:in].join(', ')}"
      end
    else
      character = options[:with].new(values)
      unless character.valid?
        record.errors.add attribute, character.errors.messages
      end
    end
  end

  def validate_each(record, attribute, values)
    case values
    when NilClass
      return nil
    when Array
      values.each do |value|
        validate_single_character(record, attribute, value, options)
      end
    else
      validate_single_character(record, attribute, values, options)
    end
  end

end


class DocumentValidator < ActiveModel::EachValidator
  def validate_single_doc(record, attribute, value, options)
    if options[:inline]
      return record.errors.add attribute, "#{value.inspect} should be an inlined doc that is converted to a hash" unless value.is_a? Hash
      record.errors.add attribute, character.errors.messages unless ThingTank.new(value).valid?
    else
      match = (value =~ /^\!\#(.+)$/)
      doc_id = $1
      return record.errors.add attribute, "#{value.inspect} does not begin with !# and is therefor no doc id" unless match      
      record.errors.add attribute, "doc with id #{doc_id} does not exist" unless record._doc.class.get(doc_id)
    end
  end

  def validate_each(record, attribute, values)
    case values
    when NilClass
      return nil
    when Array
      values.each do |value|
        validate_single_doc(record, attribute, value, options)
      end
    else
      validate_single_doc(record, attribute, values, options)
    end
  end
end